#! /usr/bin/env racket
#lang racket

;; Written by Andrew Fletcher on 17 September 2015 in Edinburgh, Scotland.

(provide read-adif)
(provide print-cpsarc-csv)

(require parser-tools/lex)
(require "maidenhead2latlong.rkt")
(require "great-circle-distance.rkt")

(require racket/date)

;; lexically analyse an ADIF file with a CPSARC CSV result

;; an ADIF file has an optional header started by a character other than '<'
;; the optional header ends '<EOH>' or '<eoh>'
;; the rest of the file is composed of records 
;; each record comprises multiple fields and ends with '<EOR>' or '<eor>'
;; each field is of the form <FIELDNAME:FIELDLENGTH>FIELDVALUE
;; e.g <call:6>WN4AZY<band:3>20M<mode:4>RTTY<qso_date:8:d>19960513<time_on:4>1305<eor>
;; whitespace outside fields should be removed
;; see <http://adif.org/304/ADIF_304.htm> for the gory details

;; defaults
(define home-loc "IO85JW")
(define antenna "Small Tx Loop")

(define-lex-abbrevs
  (digit10 (char-range "0" "9")))

;; call once only to see if a header exists
(define header?-lexer
 (lexer
   [#\< #f]
   [any-char #t]))

;; call until the end of header is found, ignoring the header
(define header-lexer 
 (lexer 
   [(union "<EOH>" "<eoh>")  #f]
   [any-char #t]))
   
(define record-lexer
  (lexer
    [(union "<EOR>" "<eor>") (begin
                               (set! qsos (cons (make-hash) qsos))
                               #t)]
    [#\< (field-lexer input-port)]
    [whitespace (record-lexer input-port)]
    [any-char (printf "*** ERROR: any-char in record-lexer is ~a ***\n" lexeme)]
    [(eof) #f]))

(define field-lexer
  (lexer
     [(concatenation (repetition 1 +inf.0 (union alphabetic numeric #\_))
                     #\: 
                     (repetition 1 2 digit10)
                     (repetition 0 1 #\:)
                     (repetition 0 +inf.0 (union alphabetic numeric #\_))
                     #\>)
      (begin
        (hash-set! (car qsos) (string-downcase(car (regexp-split #rx":" lexeme))) 
                       (field-value-lexer input-port))
        #t)]
     [whitespace (field-lexer input-port)]))

(define field-value-lexer
  (lexer
   [(concatenation (repetition 1 +inf.0 (char-complement #\<))) lexeme]))

(define (read-header in)
  (when (header-lexer in) (read-header in)))
  
(define (read-record in)
   (when (record-lexer in)
       (read-record in)))

(define (string2loc s)
  ;; quick hack
  ;; LL is near the middle of a 4 char location
  (list (cons (car (string->list (substring s 0 1))) (car (string->list (substring s 1 2))))
        (cons (string->number (substring s 2 3)) (string->number (substring s 3 4)))
        (cond [(>= (string-length s) 6)
                (cons (car (string->list (substring s 4 5))) (car (string->list (substring s 5 6))))]
              [else (cons #\L #\L)])))

(define (qsomode mode)
  ;; return mode: one of SSB/AM/FM/CW/DATA
  ;; return DATA if not one of the above
  (if (member mode (list "SSB" "AM" "FM" "CW" "DATA"))
              mode
              "DATA")
)

(define (qsosubmode mode)
    ;; return mode if it is not one of SSB/AM/FM/CW/DATA otherwise return empty string
  (if (member mode (list "SSB" "AM" "FM" "CW" "DATA"))
              ""
              mode)
)

(define (yyyymmdd->ddmmyy date)
  ;; return the date as dd/mm/yy from yyyymmdd
  (string-append (string-trim (substring date 6)) "/" (substring date 4 6) "/" (substring date 2 4))
  )

;; Given a day, month, and year, return the weekday
;; from https://stackoverflow.com/questions/13425742/racket-scheme-program-to-find-day-of-week-given-date
(define (day-month-year->weekday day month year)
  (define local-secs (find-seconds 0
                                   0
                                   0
                                   day
                                   month
                                   year
                                   #t))
  (define the-date (seconds->date local-secs))
  (vector-ref #("Sun" "Mon" "Tue" "Wed" "Thu" "Fri" "Sat")
              (date-week-day the-date)))

(define (band->mhz band)
  (cadr (assoc band (list
                     (list 70 432)
                     (list 2 144)
                     (list 50 6)
                     (list 10 28)
                     (list 12 24)
                     (list 15 21)
                     (list 17 18)
                     (list 20 14)
                     (list 30 10)
                     (list 40 7)
                     (list 80 3.5)
                     (list 160 1.8)))))
  
(define (print-record csv qso)
  (unless (eq? (hash-ref qso "call" "unknown") "unknown")
    (fprintf csv "~a,~a,~a,~a,~a,~a,~a,,~a,~a,~a,~a,~a:~a\n"
           (yyyymmdd->ddmmyy (hash-ref qso "qso_date"))
           (day-month-year->weekday (string->number (string-trim (substring (hash-ref qso "qso_date") 6)))
                                    (string->number (string-trim (substring (hash-ref qso "qso_date") 4 6)))
                                    (string->number (string-trim (substring (hash-ref qso "qso_date") 0 4))))
           (band->mhz (string->number (string-trim (hash-ref qso "band" "ERROR: bad band") "m ")))
           antenna
           (hash-ref qso "tx_pwr" "ERROR: bad tx_pwr")
           (qsomode (string-trim (hash-ref qso "mode" "ERROR: bad mode")))
           (qsosubmode (string-trim (hash-ref qso "mode" "ERROR: bad mode")))
           (hash-ref qso "call" "ERROR: bad call")
           home-loc
           (hash-ref qso "gridsquare" "ERROR: bad grid")
           (exact-round (great-circle-distance 
             (degrees->radians (car (maiden2latlong home-loc-pairs)))
             (degrees->radians (cdr (maiden2latlong home-loc-pairs)))
             (degrees->radians (car (maiden2latlong (string2loc (hash-ref qso "gridsquare" "BA00AD")))))
             (degrees->radians (cdr (maiden2latlong (string2loc (hash-ref qso "gridsquare" "BA00AD")))))))
           (substring (hash-ref qso "time_off" "ERROR: bad time") 0 2)
           (substring (hash-ref qso "time_off" "ERROR: bad time") 2 4))))
  
(define qsos (cons (make-hash) null))

(define (read-adif in)
  (call-with-input-file in
    (lambda (in)
      (if (header?-lexer in) 
        (read-header in) (unget in))
      (read-record in))))
 
(define (unget port)
  (file-position port (- (file-position port) 1)))
  
(define home-loc-pairs (string2loc home-loc))

(define (print-cpsarc-csv csv)
  (call-with-output-file csv #:exists 'replace
    (lambda (csv)
      (print-cpsarc-csv1 csv (reverse qsos)))))

(define (print-cpsarc-csv1 csv outqsos)
  (cond [(null? (cdr outqsos)) #t]
        [(begin
           (print-record csv (car outqsos))
           (print-cpsarc-csv1 csv (cdr outqsos)))]))

;; testing    
;;(read-adif "test.adif")
;;(print-cpsarc-csv "test.csv")

